/*
 *  Tcpjunk - TCP protocols testing tool
 *  Contact: oferkv@gmail.com
 *
 *  Tcpjunk is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2, or (at your option) any
 *  later version.
 *
 *  Tcpjunk is distributed in the hope that it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netdb.h>
#include <unistd.h>
#include <sys/wait.h>
#include "tcpjunk.h"
#include "utils.h"
#include "net.h"
#include "tags.h"

void client_main(struct tj_data *inst_data)
  {
  /* setup session send buffer */
  int send_buf_len = inst_data->sess_data_len;
  char *send_buf = NULL;
  send_buf = tj_malloc(send_buf, send_buf_len);

  /* other stuff */
  int buf_id;
  int loop_counter = inst_data->loop;
  int keepalive_counter = inst_data->keepalive;
  int rslt;
  int cl_socket;
  int r_timeout;
  int src_ip_cur_ses = 0;
  int bind_to;

  /* SSL Stuff */
  int ssl_err;
  SSL_CTX *ctx = NULL;
  SSL *ssl = NULL;
  const SSL_METHOD *meth;

  /* Prepare SSL Stuff if needed */
  if (inst_data->use_ssl == 1)
	{
	SSL_library_init();
	SSL_load_error_strings();

	/* Create SSL_CTX structure */
	meth = SSLv3_method();
	ctx = SSL_CTX_new(meth);
	if ((ctx) == NULL)
	  {
	  tj_log(inst_data, "Cannot create an SSL_CTX structure", c_red);
	  exit(EXIT_FAILURE);
	  }
	/* SSL structure */
	ssl = SSL_new(ctx);
	if ((ssl) == NULL)
	  {
	  tj_log(inst_data, "Cannot create an SSL structure", c_red);
	  exit(EXIT_FAILURE);
	  }
	}

  /* main client loop */
  while (inst_data->loop == 0 ? 1 : loop_counter--)
	{
	/* choose session file automatically */
	if (inst_data->auto_file > 0)
	  {
	  set_auto_session_file(inst_data);

	  /* load session file */
	  load_session_data(inst_data);

	  /* check session data */
	  find_sections(inst_data);
	  }

	/* create client socket */
	cl_socket = socket(inst_data->remote_addr->ai_family, SOCK_STREAM, 0);
	if (cl_socket < 0)
	  {
	  sprintf(log_msg, "socket allocation failed: %s", strerror(errno));
	  tj_log(inst_data, log_msg, c_red);
	  }

	/* bind client socket */
	if (inst_data->src_ip[0])
	  {
	  char bind_if_name[32];
	  char new_src_ip[128];

	  /* each IP is used for a number of sessions */
	  if (src_ip_cur_ses)
		src_ip_cur_ses--;
	  else
		{
		src_ip_cur_ses = inst_data->src_ip_num_ses - 1;

		/* IP needs to be set for this connection */
		if (inst_data->src_ip[0])
		  {
		  /* setup alias */
		  sprintf(bind_if_name, "%s:%d", inst_data->interface,
			  inst_data->inst_num + inst_data->offset - 1);

		  /* generate source IP */
		  strcpy(new_src_ip, inst_data->src_ip);
		  if (strchr(new_src_ip, 'x'))
			gen_new_src_ip(inst_data, new_src_ip);

		  /* setup IP on interface */
		  rslt = set_src_ip(inst_data, bind_if_name, new_src_ip);
		  if (rslt < 0)
			goto ip_bind_failure;
		  }
		}

	  /* clear last address */
	  if (inst_data->local_addr)
		freeaddrinfo(inst_data->local_addr);

	  /* check address */
	  inst_data->local_addr = check_address(inst_data, new_src_ip,
		  inst_data->src_port, ai_passive);

	  /* bind socket to IP address */
	  if (inst_data->verbose == 1)
		{
		sprintf(log_msg, "binding to IP: %s ", new_src_ip);
		tj_log(inst_data, log_msg, c_black);
		}
	  bind_to = 5;
	  while (bind(cl_socket, inst_data->local_addr->ai_addr,
		  inst_data->local_addr->ai_addrlen) < 0 && bind_to)
		{
		bind_to--;
		if (!bind_to)
		  {
		  sprintf(log_msg, "binding socket to IP failed: %s", strerror(errno));
		  tj_log(inst_data, log_msg, c_red);
		  }
		usleep(500000);
		}
	  }
	else if (inst_data->src_port[0])
	  {
	  /* clear last address */
	  if (inst_data->local_addr)
		freeaddrinfo(inst_data->local_addr);

	  inst_data->local_addr = check_address(inst_data, "", inst_data->src_port,
		  ai_passive);
	  if (bind(cl_socket, inst_data->local_addr->ai_addr,
		  inst_data->local_addr->ai_addrlen) < 0)
		{
		sprintf(log_msg, "binding socket to port failed: %s", strerror(errno));
		tj_log(inst_data, log_msg, c_red);
		}
	  }

	ip_bind_failure:

	/* connect */
	rslt = connect_tcp(inst_data, cl_socket);

	/* if no error, start transfers */
	if (rslt == 0)
	  {
	  if (inst_data->use_ssl == 1)
		{
		/* Assign socket to SSL structure */
		SSL_set_fd(ssl, cl_socket);

		/* SSL Handshake */
		ssl_err = SSL_connect(ssl);
		if ((ssl_err) == -1)
		  {
		  sprintf(log_msg, "SSL connect %s", ERR_error_string(ERR_get_error(),
			  NULL));
		  tj_log(inst_data, log_msg, c_red);
		  }
		}
	  send_data:

	  buf_id = 0;
	  r_timeout = 0;

	  /* get next send buffer from config */
	  while (buf_id < inst_data->sect_idx)
		{
		/* see what server has to say on start */
		if (buf_id == 0 && inst_data->wait_conn == 1)
		  r_timeout = read_reply_tcp(cl_socket, ssl, inst_data);

		if (r_timeout > 0)
		  {
		  tj_log(inst_data, "Session timeout", c_red);
		  goto close_con;
		  }

		/* prepare a temp send buffer */
		if (send_buf_len != inst_data->sects_len[buf_id])
		  {
		  send_buf_len = inst_data->sects_len[buf_id];
		  send_buf = tj_realloc(send_buf, send_buf_len);
		  }

		memset(send_buf, 0, send_buf_len);
		memcpy(send_buf, inst_data->sects[buf_id], send_buf_len);

		/* process tags */
		send_buf_len = repeat_tag(&send_buf, inst_data, send_buf_len);
		send_buf_len = instance_tag(&send_buf, inst_data, send_buf_len);
		send_buf_len = count_tag(&send_buf, inst_data, send_buf_len);
		send_buf_len = char_tag(&send_buf, inst_data, send_buf_len);
		send_buf_len = fuzz_tag(&send_buf, inst_data, send_buf_len);
		send_buf_len = system_tag(&send_buf, inst_data, send_buf_len);
		send_buf_len = encode_tag(&send_buf, inst_data, send_buf_len);
		send_buf_len = charr_tag(&send_buf, inst_data, send_buf_len);
		send_buf_len = tamper_tag(&send_buf, inst_data, send_buf_len);

		/* send data*/
		send_data_tcp(cl_socket, ssl, inst_data, send_buf, send_buf_len);

		/* see what server has to say, if we need to */
		if (!(inst_data->no_wait_last == 1 && buf_id == inst_data->sect_idx - 1))
		  r_timeout = read_reply_tcp(cl_socket, ssl, inst_data);

		if (r_timeout > 0)
		  {
		  tj_log(inst_data, "Session timeout", c_red);
		  goto close_con;
		  }

		buf_id++;
		}

	  /* if keep alive send again */
	  if (keepalive_counter > 1 && r_timeout == 0)
		{
		keepalive_counter--;
		goto send_data;
		}

	  /* close socket */
	  close_con: disconnect_tcp(cl_socket, inst_data, ssl, ctx);

	  if (inst_data->debug == 1)
		goto exit_child;
	  }
	else
	  {
	  /* connection failed */
	  sprintf(log_msg, "connection failed: %s", strerror(errno));
	  tj_log(inst_data, log_msg, c_red);
	  disconnect_tcp(cl_socket, inst_data, ssl, ctx);
	  }

	/* wait if need to */
	if (inst_data->wait > 0)
	  usleep(inst_data->wait);

	/* stop if only one instance debug mode */
	if (inst_data->debug == 1)
	  break;

	}
  /* end of main loop */
  exit_child:

  if (inst_data->use_ssl == 1)
	{
	ssl_err = SSL_shutdown(ssl);
	if ((ssl_err) == -1)
	  {
	  sprintf(log_msg, "SSL Disconnect %s", ERR_error_string(ERR_get_error(),
		  NULL));
	  tj_log(inst_data, log_msg, c_red);
	  }
	SSL_free(ssl);
	SSL_CTX_free(ctx);
	}

  free(send_buf);
  if (inst_data->gui == 0)
	exit(EXIT_SUCCESS);
  else
	return;

  }

void client_init(struct tj_data *inst_data, int insts_num)
  {
  int ci;

  for (ci = 1; ci <= insts_num; ci++)
	{
	inst_data->inst_num = ci;
	if (!fork())
	  {
	  srand((unsigned int) (time(NULL) ^ getpid()) + getppid());
	  client_main(inst_data);
	  exit(EXIT_SUCCESS);
	  }
	}
  signal(SIGINT, cleanup);
  wait(NULL);
  }
